import numpy as np

from .base import Distribution
from ..math.mittagleffler import mlf


class ColeCole(Distribution):
    """
    Functions based on Cole-Cole distribution
    """
    name = 'Cole-Cole'
    parameter = [r'\alpha']
    bounds = [(0, 1)]

    @staticmethod
    def distribution(taus, tau0, alpha):
        """
        Distribution of correlation times

        .. math::
            G(\ln\\tau) = \\frac{1}{2\pi} \\frac{\\sin(\\pi\\alpha)}{\cosh[\\alpha\ln(\\tau/\\tau_0)] + \cos(\\alpha \pi))}

        Args:
            taus:
            tau0:
            alpha:
        """
        z = np.log(taus/tau0)
        return np.sin(np.pi*alpha)/(np.cosh(z*alpha) + np.cos(alpha*np.pi))/(2*np.pi)

    @staticmethod
    def susceptibility(omega, tau, alpha):
        """
        Susceptibility

        .. math::
            \chi(\omega, \\tau, \\alpha) = \\frac{1}{1-(i\omega\\tau_0)^\\alpha}

        Args:
            omega:
            tau:
            alpha:
        """
        _tau = np.atleast_1d(tau)
        _omega = np.atleast_1d(omega)
        val = 1/(1+(1j*_omega[:, None]*_tau[None, :])**alpha)
        return np.conjugate(val).squeeze()

    @staticmethod
    def specdens(omega, tau, alpha):
        """
        Spectral density

        .. math::
            J(\omega, \\tau, \\alpha) = \\frac{\sin(\\alpha\pi/2)(\omega\\tau)^\\alpha}{\omega[1+(\omega\\tau)^{2\\alpha} + 2\\pi\cos(\\alpha\pi/2)(\omega\\tau)^\\alpha]}

        Args:
            omega:
            tau:
            alpha:
        """
        _tau = np.atleast_1d(tau)
        _omega = np.atleast_1d(omega)
        omtau = (_omega*_tau)**alpha
        return np.sin(alpha*np.pi/2) * omtau / (1 + omtau**2 + 2*np.cos(alpha*np.pi/2)*omtau) / omega

    @staticmethod
    def correlation(t, tau0, alpha):
        """
        Correlation function :math:`C(t)`

        .. math::
            C(t) = E_\\alpha[-(t/\\tau_0)^\\alpha]

        with Mittag-Leffler function :math:`E_a(x)`

        Args:
            t:
            tau0:
            alpha:
        """
        return mlf(-(t/tau0)**alpha, alpha)


if __name__ == '__main__':
    import matplotlib.pyplot as plt

    x = np.logspace(-3, 3, num=61)

    fig, ax = plt.subplots(2, 2)
    ax[0, 0].set_title('Distribution')
    ax[0, 1].set_title('Correlation func.')
    ax[1, 0].set_title('Spectral density')
    ax[1, 1].set_title('Susceptibility')

    for a in [0.4, 0.6, 0.8]:
        ax[0, 0].loglog(x, ColeCole.distribution(x, 1, a))
        ax[0, 1].semilogx(x, ColeCole.correlation(x, 1, a))
        ax[1, 0].loglog(x, ColeCole.specdens(x, 1, a))
        g = ax[1, 1].loglog(x, ColeCole.susceptibility(x, 1, a).imag)
        ax[1, 1].loglog(x, ColeCole.susceptibility(x, 1, a).real, '--', color=g[0].get_color())

    fig.tight_layout()

    plt.show()
