import numbers

import numpy as np
from scipy.special import psi, gammaincc

from .base import Distribution
from ..utils.constants import Eu


class ColeDavidson(Distribution):
    """
    Functions based on Cole-Davidson distribution
    """
    name = 'Cole-Davidson'
    parameter = [r'\gamma']
    bounds = [(0, 1)]

    @staticmethod
    def susceptibility(omega, tau, gamma):
        """
        Susceptibility

        .. math::
            \chi(\omega, \\tau, \\gamma) = \\frac{1}{(1-i\omega\\tau_0)^\\gamma}

        Args:
            omega:
            tau:
            gamma:
        """
        return (1/(1+1j*omega*tau)**gamma).conjugate()

    @staticmethod
    def specdens(omega, tau,gamma):
        """
        Spectral density

        .. math::
            J(\omega, \\tau, \\gamma) = \\frac{\sin[\\gamma\\arctan(\\omega\\tau)]}{\omega[1+(\omega\\tau)^2]^{\\gamma/2}}

        Args:
            omega:
            tau:
            gamma:
        """
        gam = gamma
        _w = np.atleast_1d(omega)
        _t = np.atleast_1d(tau)
        omtau = _w[:, None] * _t[None, :]

        ret_val = np.sin(gam*np.arctan2(omtau, 1)) / (1 + omtau**2)**(gam/2.) / _w[:, None]
        ret_val[_w == 0, :] = gam*_t

        return np.squeeze(ret_val)

    @staticmethod
    def mean(tau, gamma):
        """
        Mean tau

        .. math::
            \\langle \\tau \\rangle = \\gamma \\tau

        Args:
            tau:
            gamma:

        """

        return tau*gamma

    @staticmethod
    def logmean(tau, gamma):
        """
        Mean logarithm of tau

        .. math::
            \\langle \ln \\tau \\rangle = \ln\\tau + \psi(\gamma) + \mathrm{Eu}

        with the Euler's constant :math:`\mathrm{Eu} \\approx 0.57721567\ldots` and the digamma function

        .. math::
            \psi(x) = \\frac{d}{dx} \ln(\Gamma(x))

        Args:
            tau:
            gamma:

        References:
            R. Zorn: Logartihmic moments of relaxation time distribution. J. Chem. Phys. (2002). http://dx.doi.org/10.1063/1.1446035

        """
        return np.log(tau) + psi(gamma) + Eu

    @staticmethod
    def max(tau, gamma):
        """
        Calculate :math:`\\tau_\\text{peak}`, i.e. the inverse of the maximum frequency of the imaginary part of the susceptibility

        .. math::
            \\tau_\\text{peak} = \\frac{\\tau}{\\tan[\\pi/(2\gamma+2)]}

        Args:
            tau:
            gamma:

        References:
            R. Dı́az-Calleja: Comment on the Maximum in the Loss Permittivity for the Havriliak-Negami Equation.
            Macromolecules (2000). https://doi.org/10.1021/ma991082i

        """

        return tau/np.tan(np.pi/(2*gamma+2))

    @staticmethod
    def distribution(tau, tau0, gamma):
        """
        Distribution of correlation times

        .. math::
            G(\ln\\tau) =
               \\begin{cases}
                   \\frac{\sin(\pi\gamma)}{\pi} \\Big(\\frac{\\tau}{\\tau_0 - \\tau}\\Big)^\gamma & \\tau < \\tau_0 \\\\
                   0 & \\text{else}
               \end{cases}

        Args:
            tau:
            tau0:
            gamma:
        """

        if isinstance(tau, numbers.Number):
            return np.sin(np.pi*gamma) / np.pi * (1 / (tau0/tau - 1))**gamma if tau < tau0 else 0

        ret_val = np.zeros_like(tau)
        ret_val[tau < tau0] = np.sin(np.pi*gamma) / np.pi * (1 / (tau0/tau[tau < tau0] - 1))**gamma

        return ret_val

    @staticmethod
    def correlation(t, tau0, gamma):
        r"""
        Correlation function

        .. math::
            C(t, \tau_0, \gamma) = \frac{\Gamma(\gamma, t/\tau_0)}{\Gamma(\gamma)}

        with incomplete Gamma function

        .. math::
            \Gamma(\gamma, t/\tau_0) = \frac{1}{\Gamma(\gamma)}\int_{t/\tau_0}^\infty x^{\gamma-1}\text{e}^{-x}\,\mathrm{d}x

        Args:
            t:
            tau0:
            gamma:

        References:
            R. Hilfer: H-function representations for stretched exponential relaxation and non-Debye susceptibilities in glassy systems. Phys. Rev. E (2002) https://doi.org/10.1103/PhysRevE.65.061510

        """
        return gammaincc(gamma, t / tau0)


if __name__ == '__main__':
    import matplotlib.pyplot as plt

    x = np.logspace(-3, 3, num=61)

    fig, ax = plt.subplots(2, 2)

    for g in [0.4, 0.6, 0.8]:
        ax[0, 0].loglog(x, ColeDavidson.distribution(x, 1, g))
        ax[0, 1].semilogx(x, ColeDavidson.correlation(x, 1, g))
        ax[1, 0].loglog(x, ColeDavidson.specdens(x, 1, g))
        gr = ax[1, 1].loglog(x, ColeDavidson.susceptibility(x, 1, g).imag)
        ax[1, 1].loglog(x, ColeDavidson.susceptibility(x, 1, g).real, '--', color=gr[0].get_color())
    plt.show()
