import numpy as np
from scipy.special import psi

from .base import Distribution
from ..math.mittagleffler import mlf
from ..utils import Eu


class HavriliakNegami(Distribution):
    """
    Functions based on Cole-Davidson distribution
    """

    name = 'Havriliak-Negami'
    parameter = [r'\alpha', r'\gamma']
    bounds = [(0, 1), (0, 1)]

    @staticmethod
    def correlation(t, tau0, alpha, gamma):
        r"""
        Correlation function

        .. math::
            C(t, \tau_0, \alpha, \gamma) = 1 - \left(\frac{t}{\tau_0}\right)^{\alpha\gamma} \text{E}_{\alpha,\alpha\gamma+1}^\gamma\left[-\left(\frac{t}{\tau_0}\right)^\alpha\right]

        with incomplete Gamma function

        .. math::
            \Gamma(\gamma, t/\tau_0) = \frac{1}{\Gamma(\gamma)}\int_{t/\tau_0}^\infty x^{\gamma-1}\text{e}^{-x}\,\mathrm{d}x

        Args:
            t:
            tau0:
            alpha:
            gamma:

        References:
            R. Hilfer: H-function representations for stretched exponential relaxation and non-Debye susceptibilities in glassy systems. Phys. Rev. E (2002) https://doi.org/10.1103/PhysRevE.65.061510

        """
        return 1 - (t/tau0)**(alpha*gamma) * mlf(-(t/tau0)**alpha, alpha, alpha*gamma+1, gamma)

    @staticmethod
    def susceptibility(omega, tau, alpha, gamma):
        return np.conjugate(1/(1 + (1j*omega[:, None]*tau[None, :])**alpha)**gamma).squeeze()

    @staticmethod
    def specdens(omega, tau, alpha, gamma):
        omtau = (omega[:, None]*tau[None, :])**alpha

        zaehler = np.sin(gamma * np.arctan2(omtau*np.sin(0.5*alpha*np.pi), 1 + omtau*np.cos(0.5*alpha*np.pi)))
        nenner = (1 + 2*omtau * np.cos(0.5*alpha*np.pi) + omtau**2)**(0.5*gamma)

        return ((1 / omega) * (zaehler/nenner)).squeeze()

    @staticmethod
    def distribution(tau, tau0, alpha, gamma):
        if alpha == 1:
            from .coledavidson import ColeDavidson
            return ColeDavidson.distribution(tau, tau0, gamma)
        elif gamma == 1:
            from .colecole import ColeCole
            return ColeCole.distribution(tau, tau0, alpha)
        else:
            _y = tau/tau0
            om_y = (1 + 2*np.cos(np.pi*alpha)*(_y**alpha) + _y**(2*alpha))**(-gamma/2)
            theta_y = np.arctan2(np.sin(np.pi * alpha), _y**alpha + np.cos(np.pi*alpha))

            return np.sin(gamma*theta_y) * om_y * (_y**(alpha*gamma)) / np.pi

    @staticmethod
    def max(tau, alpha, gamma):
        return tau*(np.sin(0.5*np.pi*alpha*gamma/(gamma+1)) / np.sin(0.5*np.pi*alpha/(gamma+1)))**(1/alpha)

    @staticmethod
    def logmean(tau, alpha, gamma):
        r"""
        Calculate mean logarithm of tau

        .. math::
            \langle \ln \tau \rangle = \ln \tau + \frac{\Psi(\gamma) + \text{Eu}}{\alpha}

        Args:
            tau:
            alpha:
            gamma:

        Returns:

        """
        return np.log(tau) + (psi(gamma)+Eu)/alpha

    @staticmethod
    def mean(tau, alpha, gamma):
        # approximation according to Th. Bauer et al., J. Chem. Phys. 133, 144509 (2010).
        return tau * alpha*gamma


if __name__ == '__main__':
    import matplotlib.pyplot as plt

    x = np.logspace(-7, 3)
    y = HavriliakNegami.correlation(x, 1, 0.8, 0.26)

    plt.semilogx(x, y)
    plt.show()
