import re
import struct
import warnings
from pathlib import Path

import numpy as np

from ..data.signals import Signal

MAGIC_RE = re.compile(b'NOVOCONTROL DK-RESULTS')


class BDSReader:
    def __init__(self, fname: str = None):
        self.opts = {}
        self.fname = None
        self.x = None
        self.y = None

        if fname is not None:
            self.fname = Path(fname)
            self._parse_file()

    def __call__(self, fname: str):
        self.fname = Path(fname)
        self._parse_file()

        return self

    def _parse_file(self):
        self.opts = {}
        with self.fname.open(mode='rb') as f:
            (magic_part,) = struct.unpack('<22s', f.read(22))
            if re.match(MAGIC_RE, magic_part) is None:
                raise TypeError(f'{self.fname} not a valid file')

            f.seek(48)
            self.opts['voltage'] = struct.unpack('f', f.read(4))[0]
            self.opts['parameter'] = list(struct.unpack('i', f.read(4)))

            f.seek(2, 1)
            self.opts['edge correction'] = bool(struct.unpack('b', f.read(1))[0])
            f.seek(1, 1)
            self.opts['electrode thickness'] = struct.unpack('f', f.read(4))[0]

            self.opts['parameter'] += struct.unpack('3f', f.read(12))

            f.seek(2, 1)
            self.opts['capacity'] = struct.unpack('f', f.read(4))[0]
            self.opts['spacer area'] = struct.unpack('f', f.read(4))[0]
            f.seek(20, 1)

            self.opts['diameter'] = struct.unpack('f', f.read(4))[0]
            self.opts['area'] = self.opts['diameter']**2 * np.pi / 4
            self.opts['thickness'] = struct.unpack('f', f.read(4))[0]
            self.opts['C_0'] = (self.opts['area']-self.opts['spacer area']) / self.opts['thickness'] * 8.8541878128e-12

            f.seek(158)
            name_length = struct.unpack('b', f.read(1))[0]

            f.seek(164)
            b = f.read(name_length)
            self.opts['name'] = str(struct.unpack(f'<{name_length}s', b)[0][:-1], encoding='utf-8')

            _ = struct.unpack('h', f.read(2))
            freq_values_length = struct.unpack('b', f.read(1))[0]

            f.seek(7, 1)
            self.x = np.empty(freq_values_length)
            for freqs in range(freq_values_length):
                self.x[freqs] = struct.unpack('f', f.read(4))[0]

            _ = struct.unpack('h', f.read(2))
            length_temps = struct.unpack('b', f.read(1))[0]

            f.seek(7, 1)
            self.opts['temperatures'] = np.array(struct.unpack('f' * length_temps, f.read(4*length_temps)))

            f.seek(110, 1)
            date_length = (struct.unpack('b', f.read(1)))[0]

            f.seek(5, 1)
            self.opts['date'] = str(struct.unpack(f'<{date_length}s', f.read(date_length))[0][:-1],
                                    encoding='utf-8')

            f.seek(47, 1)
            # there are 9 different values per frequency and temperature

            _y = []
            actual_temps_length = 0
            read_length = 9 * freq_values_length
            while True:
                try:
                    _y += struct.unpack('f' * read_length, f.read(4*read_length))
                    actual_temps_length += 1
                except struct.error:
                    break

            if actual_temps_length != length_temps:
                warnings.warn('Number of set temperatures does not match number of data points')

            _y = np.array(_y).reshape((actual_temps_length, freq_values_length, 9))
            # last 3 entries are zero, save only 6
            self.y = np.transpose(_y[:, :, :6], (2, 0, 1))

    def export(self, mode='epsilon') -> list:
        if mode == 'epsilon':
            eps = self.epsilon()
        elif mode == 'sigma':
            eps = self.sigma()
        elif mode == 'modulus':
            eps = self.modulus()
        else:
            raise ValueError(f'Unknown mode {mode}, not "epsilon", "sigma", or "modulus".')

        ret_val = []

        useful_info = {
            'voltage': self.opts['voltage'],
            'diameter': self.opts['diameter'],
            'area': self.opts['area'],
            'thickness': self.opts['thickness'],
            'mode': mode
        }

        for i, temps in enumerate(self.temperature):
            ret_val.append(Signal(x=self.frequencies, y=eps[0][i, :] + 1j*eps[1][i, :],
                                  name=f'{temps:.2f}', value=temps, **useful_info))

        return ret_val

    def __getitem__(self, val):
        return self.opts[val]

    @property
    def set_temp(self):
        return self.opts['temperatures'][:len(self.temperature)]

    @property
    def probe_temp(self):
        return self.y[2]

    @property
    def temperature(self):
        return self.probe_temp.mean(axis=1)

    @property
    def frequencies(self):
        return self.x

    def capacity(self):
        # NOTE: nobody ever used edge correction, so we ignore this option entirely
        return self.y[0] - self.opts['capacity'], -1./self.y[1] / (self.frequencies*2*np.pi)

    def voltage(self):
        return self.y[5]

    def sigma(self):
        # sigma^* = i * omega * eps_0 * (eps^* - 1)
        eps_r, eps_i = self.epsilon()
        conv = 0.01 * 8.8541878128e-12 * 2*np.pi*self.x  # factor 0.01: sigma in S/cm
        return conv * eps_i, conv * (eps_r-1)

    def epsilon(self):
        # eps^* = Cp^* d/A/eps_0
        cr, ci = self.capacity()
        return cr / self.opts['C_0'], -ci / self.opts['C_0']

    def modulus(self):
        # M^* = 1/eps^*
        eps_r, eps_i = self.epsilon()
        magn = eps_r ** 2 + eps_i ** 2
        return eps_r / magn, eps_i / magn

    def loss_factor(self):
        eps_r, eps_i = self.epsilon()

        return eps_i / eps_r, None
