from math import log
from collections import OrderedDict

from ..utils.constants import pi, gamma, mu0, hbar


__all__ = ['Quadrupolar', 'QuadrupolarQCC', 'Czjzek', 'HeteroDipolar', 'HomoDipolar', 'Constant']


class Coupling(object):
    name = 'coupling'
    parameter = None
    choice = None
    unit = None
    equation = ''


class Quadrupolar(Coupling):
    name = 'Quadrupolar'
    parameter = [r'\delta', r'\eta']
    unit = ['Hz', '']
    equation = r'24 (2I-1)(2I+3) / (25(6m+3)^{2}) (1+\eta^2/3) \pi^{2} \delta^{2}'
    choice = [('Spin', 'spin', OrderedDict([('1', 1), ('3/2', 1.5), ('2', 2),
                                            ('5/2', 2.5), ('3', 3), ('7/2', 3.5)]))]

    @staticmethod
    def func(delta, eta, spin=1):
        m = 0 if (2*spin) % 2 == 0 else 0.5
        return 24*(2*spin-1)*(2*spin+3) * pi**2 * delta**2 * (1+eta**2 / 3) / 25 / (6*m+3)**2


class QuadrupolarQCC(Coupling):
    name = 'Quadrupolar (QCC)'
    parameter = ['C_{Q}', r'\eta']
    unit = ['Hz', '']
    choice = [('Spin', 'spin', {'1': 1, '3/2': 1.5, '5/2': 2.5, '7/2': 3.5})]
    equation = r'3/50 \pi^{2} C_{Q}^{2} (1+\eta^{2}/3) (2I+3)/(I^{2}(2I-1))'

    @staticmethod
    def func(cq, eta, spin=1):
        return 0.06 * (2*spin+3)*pi**2 * cq**2 * (1+eta**2 / 3) / spin**2 / (2*spin-1)


class Czjzek(Coupling):
    name = 'Czjzek'
    parameter = [r'\Sigma']
    unit = ['Hz']
    choice = [('Spin', 'spin', {'1': 1, '3/2': 1.5, '5/2': 2.5, '7/2': 3.5})]
    equation = r'3/[20log(2)] \pi^{2} \Sigma^{2} (2I+3)/(I^{2}(2I-1))'

    @staticmethod
    def func(fwhm, spin=1):
        return 3*(2+spin+3) * pi**2 * fwhm**2 / 20 / log(2) / spin**2 / (2*spin-1)


class HeteroDipolar(Coupling):
    name = 'Heteronuclear Dipolar'
    parameter = ['r']
    unit = ['m']
    choice = [(r'\gamma_{1}', 'nucleus1', {k: k for k in gamma.keys()}),
              (r'\gamma_{2}', 'nucleus2', {k: k for k in gamma.keys()})]
    equation = r'1/10 * [\mu_{0}/(4\pi) * \hbar * \gamma_{1}\gamma_{2}/r^{3}]^{2}'

    @staticmethod
    def func(r, nucleus1='1H', nucleus2='19F'):
        if nucleus1 not in gamma:
            raise KeyError('Unknown nucleus {}'.format(nucleus1))
        if nucleus2 not in gamma:
            raise KeyError('Unknown nucleus {}'.format(nucleus2))

        coupling = mu0 / (4*pi) * hbar*gamma[nucleus1]*gamma[nucleus2] / (r+1e-34)**3
        coupling **= 2

        return 0.1 * coupling


class HomoDipolar(Coupling):
    name = 'Homonuclear Dipolar'
    parameter = ['r']
    unit = ['m']
    choice = [(r'\gamma', 'nucleus', {k: k for k in gamma.keys()})]
    equation = r'3/10 * [\mu_{0}/(4\pi) * \hbar * \gamma^{2}/r^{3}]^{2}'

    @staticmethod
    def func(r, nucleus='1H'):
        if nucleus not in gamma:
            raise KeyError('Unknown nucleus {}'.format(nucleus))

        coupling = mu0 / (4*pi) * hbar * gamma[nucleus]**2 / (r+1e-34)**3
        coupling **= 2

        return 0.3 * coupling


class Constant(Coupling):
    name = 'Constant'
    parameter = ['C']
    equation = 'C'

    @staticmethod
    def func(c):
        return c
