from numbers import Number
from pathlib import Path
from typing import Tuple
from warnings import warn

import numpy as np
from scipy.interpolate import interp1d, Akima1DInterpolator
from scipy.optimize import minimize

from nmreval.lib.utils import ArrayLike
from ..distributions.debye import Debye as Debye


class Relaxation:
    def __init__(self, distribution=None):
        self._distribution = distribution
        self._dist_parameter = ()
        self._dist_kw = {}

        self._coupling = None
        self._coup_parameter = ()
        self._coup_kw = {}

        self._prefactor = 1.

    def __repr__(self):
        if self._distribution is not None:
            return str(self._distribution.name)
        else:
            return super().__repr__()

    def coupling(self, coupling, parameter: list = None, keywords: dict = None):
        self._coupling = coupling

        if parameter is not None:
            self._coup_parameter = parameter

        if keywords is not None:
            self._coup_kw = keywords

        if isinstance(self._coupling, Number):
            self._prefactor = self._coupling
        else:
            try:
                self._prefactor = self._coupling.func(*self._coup_parameter, **self._coup_kw)
            except TypeError:
                pass

    def distribution(self, dist, parameter=None, keywords=None):
        self._distribution = dist

        if parameter is not None:
            self._dist_parameter = parameter

        if keywords is not None:
            self._dist_kw = keywords

    def t1(self, omega, tau, *args, mode='bpp', **kwargs):
        # defaults to BPP
        if mode == 'bpp':
            return self.t1_bpp(omega, tau, *args, **kwargs)
        elif mode == 'dipolar':
            return self.t1_dipolar(omega, tau, *args, **kwargs)
        else:
            raise AttributeError(f'Unknown mode {mode}. Use either "bpp" or "dipolar".')

    def t1_dipolar(self, omega: ArrayLike, tau: ArrayLike, *args,
                   inverse: bool = True, prefactor: float = None, **kwargs) -> ArrayLike:
        """

        Args:
            omega:
            tau:
            *args:
            inverse:
            prefactor:
            **kwargs:

        Returns:

        """
        try:
            omega_2 = kwargs['omega_coup']
        except KeyError:
            if kwargs.get('gamma_obs', False):
                omega_2 = kwargs['gamma_coup'] / kwargs['gamma_obs'] * omega
            else:
                raise AttributeError('Unknown second frequency.')

        if prefactor is None:
            prefactor = self._prefactor

        if len(args) == 0:
            args = self._dist_parameter

        rate = prefactor * (3 * self._distribution.specdens(omega, tau, *args) +
                            self._distribution.specdens(np.abs(omega - omega_2), tau, *args) +
                            6 * self._distribution.specdens(omega + omega_2, tau, *args))
        if inverse:
            return 1 / rate
        else:
            return rate

    def t1_bpp(self, omega, tau, *args, inverse=True, prefactor=None, **kwargs):
        if prefactor is None:
            prefactor = self._prefactor

        if len(args) == 0:
            args = self._dist_parameter

        rate = prefactor * (self._distribution.specdens(omega, tau, *args) +
                            4*self._distribution.specdens(2*omega, tau, *args))
        if inverse:
            return 1. / rate
        else:
            return rate

    def t1_rho(self, omega, tau, omega_rf, *args, inverse=True, prefactor=None, **kwargs):
        if prefactor is None:
            prefactor = self._prefactor

        if len(args) == 0:
            args = self._dist_parameter

        rate = prefactor * (10 * self._distribution.specdens(omega, tau, *args) +
                            4 * self._distribution.specdens(2 * omega, tau, *args) +
                            6 * self._distribution.specdens(omega_rf, tau, *args))
        if inverse:
            return 1. / rate
        else:
            return rate

    def t2(self, omega, tau, *args, inverse=True, prefactor=None, **kwargs):
        if prefactor is None:
            prefactor = self._prefactor / 2.

        if len(args) == 0:
            args = self._dist_parameter

        rate = prefactor * (3 * self._distribution.specdens(0, tau, *args) +
                            5 * self._distribution.specdens(omega, tau, *args) +
                            2 * self._distribution.specdens(2 * omega, tau, *args))
        if inverse:
            return 1. / rate
        else:
            return rate


class RelaxationEvaluation(Relaxation):
    def __init__(self, distribution=None):
        super().__init__(distribution=distribution)
        self.t1min = (np.nan, np.nan)
        self._interpolate = None
        self._interpolate_range = (None, None)

        self.omega = np.nan
        self.x = None
        self.y = None

    def data(self, temp, t1):
        temp = np.asarray(temp)
        t1 = np.asarray(t1)
        sortidx = temp.argsort()
        self.x = temp[sortidx]
        self.y = t1[sortidx]
        self.calculate_t1_min()

    def calculate_t1_min(self, interpolate: int = None, trange=None):
        min_index = np.argmin(self.y)
        t1_min = (self.x[min_index], self.y[min_index])
        parabola = None
        self._interpolate_range = (None, None)

        if interpolate is not None:
            if interpolate not in [0, 1, 2, 3]:
                raise ValueError(f'Unknown interpolation value {interpolate}')

            if interpolate != 0:
                if trange is None:
                    left_b = max(min_index-2, 0)
                    right_b = min(len(self.x), min_index+3)
                else:
                    left_b = np.argmin(np.abs(self.x-trange[0]))
                    right_b = np.argmin(np.abs(self.x-trange[1]))+1

                temp_range = self.x[left_b:right_b]
                t1_range = self.y[left_b:right_b]

                try:
                    x_inter = np.linspace(temp_range[0], temp_range[-1], num=51)

                    if interpolate == 1:
                        i_func = np.poly1d(np.polyfit(temp_range, t1_range, 2))
                    elif interpolate == 2:
                        i_func = interp1d(temp_range, t1_range, kind='cubic')
                    else:
                        i_func = Akima1DInterpolator(temp_range, t1_range)

                    y_inter = i_func(x_inter)
                    t1_min = (x_inter[np.argmin(y_inter)], y_inter[np.argmin(y_inter)])
                    self._interpolate = i_func
                    parabola = (x_inter, y_inter)
                    self._interpolate_range = (temp_range[0], temp_range[-1])

                except IndexError:
                    pass

        self.t1min = t1_min

        return t1_min, parabola

    def increase(self, variable: Tuple[str, int], height: float = None, omega: float = None,
                 dist_parameter: list = None, prefactor=None,
                 coupling_kwargs=None):

        stretching = mini = np.nan
        if height is None:
            height = self.t1min[1]

        if omega is None:
            if np.isnan(self.omega):
                raise ValueError('No larmor frequency set')

            omega = self.omega

        if prefactor is None:
            prefactor = self._prefactor

        if dist_parameter is None:
            dist_parameter = self._dist_parameter

        tau_lims = np.log10(1/omega)-3, np.log10(1/omega)+3

        if not variable:
            if isinstance(prefactor, list):
                if coupling_kwargs is None:
                    coupling_kwargs = self._coup_kw
                prefactor = self._coupling.func(*prefactor, **coupling_kwargs)

            return stretching, np.min(self.t1(omega, np.logspace(*tau_lims, num=1001),
                                              *dist_parameter, prefactor=prefactor, inverse=True))

        if variable[0] == 'distribution':
            if isinstance(self._distribution, Debye):
                return 1.

            if isinstance(prefactor, list):
                if coupling_kwargs is None:
                    coupling_kwargs = self._coup_kw
                prefactor = self._coupling.func(*prefactor, **coupling_kwargs)

            use_fmin = True
            if self._distribution.name in ['KWW', 'Cole-Davidson', 'Cole-Cole', 'Log-Gaussian']:

                from importlib.resources import path
                with path('resources.nmr', 'T1_min.npz') as fp:
                    if fp.exists():
                        t1min_table = np.load(str(fp))[self._distribution.name.replace('-', '_')]
                        debye_min = omega / 1.42517571908650 / prefactor  # Zahl kommt von Wolfram alpha
                        ratio = height / debye_min
                        stretching = t1min_table[np.argmin(np.abs(t1min_table[:, 1]-ratio)), 0]
                        use_fmin = False

            if use_fmin:
                def _f(_x, p, ii):
                    p[ii] = _x[0]
                    t1_calc = np.min(self.t1(omega, np.logspace(*tau_lims, num=1001), *p,
                                             prefactor=prefactor, inverse=True))

                    return np.abs(t1_calc - height)

                stretching = minimize(_f, np.array([dist_parameter[variable[1]]]),
                                      args=(dist_parameter, variable[1]),
                                      method='Nelder-Mead').x[0]

        elif variable[0] == 'coupling':
            t1_no_coup = np.min(self.t1(omega, np.logspace(*tau_lims, num=1001),
                                        *dist_parameter, prefactor=1))

            if isinstance(prefactor, list):
                def _f(_x, p, ii):
                    p[ii] = _x
                    return np.abs(t1_no_coup/height - self._coupling.func(*p, **coupling_kwargs)[0])

                stretching = minimize(_f, np.array([prefactor[variable[1]]]),
                                      args=(prefactor, variable[1])).x[0]
            else:
                stretching = t1_no_coup / height

        else:
            raise ValueError('Use "distribution" or "coupling" to set parameter')

        if stretching:
            self._prefactor = prefactor
            self._dist_parameter = dist_parameter
            if isinstance(prefactor, list):
                self._coup_parameter = prefactor
            mini = np.min(self.t1(omega, np.logspace(*tau_lims, num=1001), *self._dist_parameter,
                                  prefactor=self._prefactor))

        return stretching, mini

    def correlation_from_t1(self, mode: str = 'raw', interpolate=False, omega=None,
                            dist_parameter: list = None, prefactor: float = None,
                            coupling_param: list = None, coupling_kwargs: dict = None):

        if self.x is None:
            raise ValueError('Temperature is not set')

        if self.y is None:
            raise ValueError('T1 data is not set')

        if omega is None:
            if np.isnan(self.omega):
                raise ValueError('No frequency set')
            omega = self.omega

        if prefactor is None:
            if coupling_kwargs is None:
                coupling_kwargs = self._coup_kw

            if coupling_param is None:
                prefactor = self._prefactor
            else:
                prefactor = self._coupling.func(*coupling_param, **coupling_kwargs)

        if dist_parameter is None:
            dist_parameter = self._dist_parameter

        fast_idx = self.x > self.t1min[0]

        slow_t1 = self.y[~fast_idx]
        slow_temp = self.x[~fast_idx]

        correlation_times = np.ones(self.x.shape)
        offset = len(slow_t1)

        base_taus = np.logspace(-10, -7, num=1001)
        min_tau = base_taus[np.argmin(self.t1(omega, base_taus, *dist_parameter, prefactor=prefactor))]

        taus = np.geomspace(min_tau, 100.*min_tau, num=501)
        current_t1 = self.t1(omega, taus, *dist_parameter, prefactor=prefactor)

        for i in range(1, len(slow_t1)+1):
            t1_i = slow_t1[-i]

            if interpolate and self._interpolate is not None:
                if slow_temp[-i] >= self._interpolate_range[0]:
                    t1_i = self._interpolate(slow_temp[-i])

            if min(current_t1) > t1_i:
                warn('Correlation time could not be calculated')
                correlation_times[offset-i] = taus[0]
                continue

            cross_idx = np.where(np.diff(np.sign(current_t1 - t1_i)))[0]
            while len(cross_idx) == 0:
                taus *= 10
                current_t1 = self.t1(omega, taus, *dist_parameter, prefactor=prefactor)
                cross_idx = np.where(np.diff(np.sign(current_t1 - t1_i)))[0]

            lamb = (t1_i - current_t1[cross_idx]) / (current_t1[cross_idx+1]-current_t1[cross_idx])
            correlation_times[offset-i] = (taus[cross_idx+1] * lamb + (1 - lamb) * taus[cross_idx])[0]

        fast_t1 = self.y[fast_idx]
        fast_temp = self.x[fast_idx]

        taus = np.geomspace(0.01*min_tau, min_tau, num=501)
        current_t1 = self.t1(omega, taus, *dist_parameter, prefactor=prefactor)

        for i in range(len(fast_t1)):
            t1_i = fast_t1[i]

            if interpolate and self._interpolate is not None:
                if fast_temp[i] <= self._interpolate_range[1]:
                    t1_i = self._interpolate(fast_temp[i])

            if current_t1[-1] > t1_i:
                correlation_times[offset+i] = taus[-1]
                warn('Correlation time could not be calculated')
                continue

            cross_idx = np.where(np.diff(np.sign(current_t1 - t1_i)))[0]
            while len(cross_idx) == 0:
                taus /= 10
                current_t1 = self.t1(omega, taus, *dist_parameter, prefactor=prefactor)
                cross_idx = np.where(np.diff(np.sign(current_t1 - t1_i)))[0]

            lamb = (t1_i - current_t1[cross_idx]) / (current_t1[cross_idx+1]-current_t1[cross_idx])
            correlation_times[offset+i] = (taus[cross_idx+1] * lamb + (1-lamb) * taus[cross_idx])[0]

        opts = {'distribution': (self._distribution.name, dist_parameter),
                'coupling': (self._coupling.name, coupling_param, coupling_kwargs),
                'frequency': omega/2/np.pi}

        return np.c_[self.x, self._distribution.mean_value(correlation_times, *dist_parameter, mode=mode)], opts

    def _create_minimum_file(self, filename):
        x = np.geomspace(0.1, 1, num=10001)
        with Path(filename).open('w') as f:
            f.write('#broadening\tT1_min/min_Debye\n')
            for i, a in enumerate(np.geomspace(0.1, 10, num=10000)):
                alpha = a
                t1_i = self.t1_bpp(1, x, alpha)
                f.write(f'{alpha}\t{min(t1_i) / 0.701667864144}\n')
                f.flush()
